@@ -1,25 +1,21 @@
+# frozen_string_literal: false
 =begin
-= $RCSfile$ -- Loader for all OpenSSL C-space and Ruby-space definitions
-
 = Info
   'OpenSSL for Ruby 2' project
   Copyright (C) 2002  Michal Rokos <m.rokos@sh.cvut.cz>
   All rights reserved.
 
 = Licence
-  This program is licenced under the same licence as Ruby.
+  This program is licensed under the same licence as Ruby.
   (See the file 'LICENCE'.)
-
-= Version
-  $Id: openssl.rb 12496 2007-06-08 15:02:04Z technorama $
 =end
 
 require 'openssl.so'
 
 require 'openssl/bn'
+require 'openssl/pkey'
 require 'openssl/cipher'
+require 'openssl/config'
 require 'openssl/digest'
-require 'openssl/pkcs7'
-require 'openssl/ssl'
 require 'openssl/x509'
-
+require 'openssl/ssl'
