@@ -2,7 +2,7 @@
 
   object.c -
 
-  $Author: nagachika $
+  $Author: naruse $
   created at: Thu Jul 15 12:01:24 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -11,17 +11,15 @@
 
 **********************************************************************/
 
-#include "ruby/ruby.h"
+#include "internal.h"
 #include "ruby/st.h"
 #include "ruby/util.h"
-#include "ruby/encoding.h"
 #include <stdio.h>
 #include <errno.h>
 #include <ctype.h>
 #include <math.h>
 #include <float.h>
 #include "constant.h"
-#include "internal.h"
 #include "id.h"
 #include "probes.h"
 
@@ -72,7 +70,6 @@ rb_obj_setup(VALUE obj, VALUE klass, VALUE type)
 {
     RBASIC(obj)->flags = type;
     RBASIC_SET_CLASS(obj, klass);
-    if (rb_safe_level() >= 3) FL_SET((obj), FL_TAINT);
     return obj;
 }
 
@@ -145,7 +142,11 @@ rb_obj_equal(VALUE obj1, VALUE obj2)
     return Qfalse;
 }
 
+#if 0
 /*
+ * call-seq:
+ *    obj.hash    -> fixnum
+ *
  * Generates a Fixnum hash value for this object.  This function must have the
  * property that <code>a.eql?(b)</code> implies <code>a.hash == b.hash</code>.
  *
@@ -161,7 +162,6 @@ rb_obj_equal(VALUE obj1, VALUE obj2)
 VALUE
 rb_obj_hash(VALUE obj)
 {
-    long rb_objid_hash(st_index_t index);
     VALUE oid = rb_obj_id(obj);
 #if SIZEOF_LONG == SIZEOF_VOIDP
     st_index_t index = NUM2LONG(oid);
@@ -172,6 +172,9 @@ rb_obj_hash(VALUE obj)
 #endif
     return LONG2FIX(rb_objid_hash(index));
 }
+#else
+VALUE rb_obj_hash(VALUE obj);
+#endif
 
 /*
  *  call-seq:
@@ -286,26 +289,11 @@ init_copy(VALUE dest, VALUE obj)
     }
     RBASIC(dest)->flags &= ~(T_MASK|FL_EXIVAR);
     RBASIC(dest)->flags |= RBASIC(obj)->flags & (T_MASK|FL_EXIVAR|FL_TAINT);
+    rb_copy_wb_protected_attribute(dest, obj);
     rb_copy_generic_ivar(dest, obj);
     rb_gc_copy_finalizer(dest, obj);
-    switch (TYPE(obj)) {
-      case T_OBJECT:
+    if (RB_TYPE_P(obj, T_OBJECT)) {
 	rb_obj_copy_ivar(dest, obj);
-        break;
-      case T_CLASS:
-      case T_MODULE:
-	if (RCLASS_IV_TBL(dest)) {
-	    st_free_table(RCLASS_IV_TBL(dest));
-	    RCLASS_IV_TBL(dest) = 0;
-	}
-	if (RCLASS_CONST_TBL(dest)) {
-	    rb_free_const_table(RCLASS_CONST_TBL(dest));
-	    RCLASS_CONST_TBL(dest) = 0;
-	}
-	if (RCLASS_IV_TBL(obj)) {
-	    RCLASS_IV_TBL(dest) = rb_st_copy(dest, RCLASS_IV_TBL(obj));
-	}
-        break;
     }
 }
 
@@ -343,8 +331,8 @@ rb_obj_clone(VALUE obj)
         rb_raise(rb_eTypeError, "can't clone %s", rb_obj_classname(obj));
     }
     clone = rb_obj_alloc(rb_obj_class(obj));
-    RBASIC(clone)->flags &= (FL_TAINT|FL_PROMOTED|FL_WB_PROTECTED);
-    RBASIC(clone)->flags |= RBASIC(obj)->flags & ~(FL_PROMOTED|FL_FREEZE|FL_FINALIZE|FL_WB_PROTECTED);
+    RBASIC(clone)->flags &= (FL_TAINT|FL_PROMOTED0|FL_PROMOTED1);
+    RBASIC(clone)->flags |= RBASIC(obj)->flags & ~(FL_PROMOTED0|FL_PROMOTED1|FL_FREEZE|FL_FINALIZE);
 
     singleton = rb_singleton_class_clone_and_attach(obj, clone);
     RBASIC_SET_CLASS(clone, singleton);
@@ -417,6 +405,23 @@ rb_obj_dup(VALUE obj)
     return dup;
 }
 
+/*
+ *  call-seq:
+ *     obj.itself -> an_object
+ *
+ *  Returns <i>obj</i>.
+ *
+ *	string = 'my string' #=> "my string"
+ *	string.itself.object_id == string.object_id #=> true
+ *
+ */
+
+static VALUE
+rb_obj_itself(VALUE obj)
+{
+    return obj;
+}
+
 /* :nodoc: */
 VALUE
 rb_obj_init_copy(VALUE obj, VALUE orig)
@@ -460,24 +465,26 @@ rb_any_to_s(VALUE obj)
     return str;
 }
 
+VALUE rb_str_escape(VALUE str);
 /*
- * If the default external encoding is ASCII compatible, the encoding of
- * the inspected result must be compatible with it.
- * If the default external encoding is ASCII incompatible,
+ * If the default internal or external encoding is ASCII compatible,
+ * the encoding of the inspected result must be compatible with it.
+ * If the default internal or external encoding is ASCII incompatible,
  * the result must be ASCII only.
  */
 VALUE
 rb_inspect(VALUE obj)
 {
-    VALUE str = rb_obj_as_string(rb_funcall(obj, id_inspect, 0, 0));
-    rb_encoding *ext = rb_default_external_encoding();
-    if (!rb_enc_asciicompat(ext)) {
+    VALUE str = rb_obj_as_string(rb_funcallv(obj, id_inspect, 0, 0));
+    rb_encoding *enc = rb_default_internal_encoding();
+    if (enc == NULL) enc = rb_default_external_encoding();
+    if (!rb_enc_asciicompat(enc)) {
 	if (!rb_enc_str_asciionly_p(str))
-	    rb_raise(rb_eEncCompatError, "inspected result must be ASCII only if default external encoding is ASCII incompatible");
+	    return rb_str_escape(str);
 	return str;
     }
-    if (rb_enc_get(str) != ext && !rb_enc_str_asciionly_p(str))
-	rb_raise(rb_eEncCompatError, "inspected result must be ASCII only or use the default external encoding");
+    if (rb_enc_get(str) != enc && !rb_enc_str_asciionly_p(str))
+	return rb_str_escape(str);
     return str;
 }
 
@@ -487,8 +494,6 @@ inspect_i(st_data_t k, st_data_t v, st_data_t a)
     ID id = (ID)k;
     VALUE value = (VALUE)v;
     VALUE str = (VALUE)a;
-    VALUE str2;
-    const char *ivname;
 
     /* need not to show internal data */
     if (CLASS_OF(value) == 0) return ST_CONTINUE;
@@ -500,12 +505,8 @@ inspect_i(st_data_t k, st_data_t v, st_data_t a)
     else {
 	rb_str_cat2(str, ", ");
     }
-    ivname = rb_id2name(id);
-    rb_str_cat2(str, ivname);
-    rb_str_cat2(str, "=");
-    str2 = rb_inspect(value);
-    rb_str_append(str, str2);
-    OBJ_INFECT(str, str2);
+    rb_str_catf(str, "%"PRIsVALUE"=%+"PRIsVALUE,
+		rb_id2str(id), value);
 
     return ST_CONTINUE;
 }
@@ -655,7 +656,7 @@ static VALUE
 class_search_ancestor(VALUE cl, VALUE c)
 {
     while (cl) {
-	if (cl == c || RCLASS_M_TBL_WRAPPER(cl) == RCLASS_M_TBL_WRAPPER(c))
+	if (cl == c || RCLASS_M_TBL(cl) == RCLASS_M_TBL(c))
 	    return cl;
 	cl = RCLASS_SUPER(cl);
     }
@@ -958,15 +959,12 @@ rb_obj_tainted(VALUE obj)
  *
  *  You should only untaint a tainted object if your code has inspected it and
  *  determined that it is safe. To do so use #untaint.
- *
- *  In $SAFE level 3, all newly created objects are tainted and you can't untaint
- *  objects.
  */
 
 VALUE
 rb_obj_taint(VALUE obj)
 {
-    if (!OBJ_TAINTED(obj)) {
+    if (!OBJ_TAINTED(obj) && OBJ_TAINTABLE(obj)) {
 	rb_check_frozen(obj);
 	OBJ_TAINT(obj);
     }
@@ -986,7 +984,6 @@ rb_obj_taint(VALUE obj)
 VALUE
 rb_obj_untaint(VALUE obj)
 {
-    rb_secure(3);
     if (OBJ_TAINTED(obj)) {
 	rb_check_frozen(obj);
 	FL_UNSET(obj, FL_TAINT);
@@ -1043,8 +1040,6 @@ rb_obj_infect(VALUE obj1, VALUE obj2)
     OBJ_INFECT(obj1, obj2);
 }
 
-static st_table *immediate_frozen_tbl = 0;
-
 /*
  *  call-seq:
  *     obj.freeze    -> obj
@@ -1062,8 +1057,11 @@ static st_table *immediate_frozen_tbl = 0;
  *
  *  <em>produces:</em>
  *
- *     prog.rb:3:in `<<': can't modify frozen array (RuntimeError)
+ *     prog.rb:3:in `<<': can't modify frozen Array (RuntimeError)
  *     	from prog.rb:3
+ *
+ *  Objects of the following classes are always frozen: Fixnum,
+ *  Bignum, Float, Symbol.
  */
 
 VALUE
@@ -1072,10 +1070,7 @@ rb_obj_freeze(VALUE obj)
     if (!OBJ_FROZEN(obj)) {
 	OBJ_FREEZE(obj);
 	if (SPECIAL_CONST_P(obj)) {
-	    if (!immediate_frozen_tbl) {
-		immediate_frozen_tbl = st_init_numtable();
-	    }
-	    st_insert(immediate_frozen_tbl, obj, (st_data_t)Qtrue);
+	    rb_bug("special consts should be frozen.");
 	}
     }
     return obj;
@@ -1095,12 +1090,7 @@ rb_obj_freeze(VALUE obj)
 VALUE
 rb_obj_frozen_p(VALUE obj)
 {
-    if (OBJ_FROZEN(obj)) return Qtrue;
-    if (SPECIAL_CONST_P(obj)) {
-	if (!immediate_frozen_tbl) return Qfalse;
-	if (st_lookup(immediate_frozen_tbl, obj, 0)) return Qtrue;
-    }
-    return Qfalse;
+    return OBJ_FROZEN(obj) ? Qtrue : Qfalse;
 }
 
 
@@ -1421,16 +1411,16 @@ rb_obj_not_match(VALUE obj1, VALUE obj2)
  *  Returns 0 if +obj+ and +other+ are the same object
  *  or <code>obj == other</code>, otherwise nil.
  *
- *  The <=> is used by various methods to compare objects, for example
+ *  The <code><=></code> is used by various methods to compare objects, for example
  *  Enumerable#sort, Enumerable#max etc.
  *
- *  Your implementation of <=> should return one of the following values: -1, 0,
+ *  Your implementation of <code><=></code> should return one of the following values: -1, 0,
  *  1 or nil. -1 means self is smaller than other. 0 means self is equal to other.
  *  1 means self is bigger than other. Nil means the two values could not be
  *  compared.
  *
- *  When you define <=>, you can include Comparable to gain the methods <=, <,
- *  ==, >=, > and between?.
+ *  When you define <code><=></code>, you can include Comparable to gain the methods
+ *  <code><=</code>, <code><</code>, <code>==</code>, <code>>=</code>, <code>></code> and <code>between?</code>.
  */
 static VALUE
 rb_obj_cmp(VALUE obj1, VALUE obj2)
@@ -1533,9 +1523,9 @@ rb_mod_freeze(VALUE mod)
  *     mod === obj    -> true or false
  *
  *  Case Equality---Returns <code>true</code> if <i>obj</i> is an
- *  instance of <i>mod</i> or one of <i>mod</i>'s descendants. Of
- *  limited use for modules, but can be used in <code>case</code>
- *  statements to classify objects by class.
+ *  instance of <i>mod</i> or and instance of one of <i>mod</i>'s descendants.
+ *  Of limited use for modules, but can be used in <code>case</code> statements
+ *  to classify objects by class.
  */
 
 static VALUE
@@ -1641,7 +1631,7 @@ rb_mod_gt(VALUE mod, VALUE arg)
  *
  *  Comparison---Returns -1, 0, +1 or nil depending on whether +module+
  *  includes +other_module+, they are the same, or if +module+ is included by
- *  +other_module+. This is the basis for the tests in Comparable.
+ *  +other_module+.
  *
  *  Returns +nil+ if +module+ has no relationship with +other_module+, if
  *  +other_module+ is not a module, or if the two values are incomparable.
@@ -1715,6 +1705,17 @@ rb_mod_initialize(VALUE module)
     return Qnil;
 }
 
+/* :nodoc: */
+static VALUE
+rb_mod_initialize_clone(VALUE clone, VALUE orig)
+{
+    VALUE ret;
+    ret = rb_obj_init_dup_clone(clone, orig);
+    if (OBJ_FROZEN(orig))
+        rb_class_name(clone);
+    return ret;
+}
+
 /*
  *  call-seq:
  *     Class.new(super_class=Object)               -> a_class
@@ -1771,6 +1772,13 @@ rb_class_initialize(int argc, VALUE *argv, VALUE klass)
     return klass;
 }
 
+void
+rb_undefined_alloc(VALUE klass)
+{
+    rb_raise(rb_eTypeError, "allocator undefined for %"PRIsVALUE,
+	     klass);
+}
+
 /*
  *  call-seq:
  *     class.allocate()   ->   obj
@@ -1807,18 +1815,10 @@ rb_obj_alloc(VALUE klass)
     }
     allocator = rb_get_alloc_func(klass);
     if (!allocator) {
-	rb_raise(rb_eTypeError, "allocator undefined for %"PRIsVALUE,
-		 klass);
+	rb_undefined_alloc(klass);
     }
 
-#if !defined(DTRACE_PROBES_DISABLED) || !DTRACE_PROBES_DISABLED
-    if (RUBY_DTRACE_OBJECT_CREATE_ENABLED()) {
-        const char * file = rb_sourcefile();
-        RUBY_DTRACE_OBJECT_CREATE(rb_class2name(klass),
-				  file ? file : "",
-				  rb_sourceline());
-    }
-#endif
+    RUBY_DTRACE_CREATE_HOOK(OBJECT, rb_class2name(klass));
 
     obj = (*allocator)(klass);
 
@@ -1848,7 +1848,7 @@ rb_class_allocate_instance(VALUE klass)
  */
 
 VALUE
-rb_class_new_instance(int argc, VALUE *argv, VALUE klass)
+rb_class_new_instance(int argc, const VALUE *argv, VALUE klass)
 {
     VALUE obj;
 
@@ -1901,59 +1901,45 @@ rb_class_get_superclass(VALUE klass)
     return RCLASS(klass)->super;
 }
 
-#define id_for_setter(name, type, message) \
-    check_setter_id(name, rb_is_##type##_id, rb_is_##type##_name, message)
+#define id_for_var(obj, name, part, type) \
+    id_for_setter(obj, name, type, "`%1$s' is not allowed as "#part" "#type" variable name")
+#define id_for_setter(obj, name, type, message) \
+    check_setter_id(obj, &(name), rb_is_##type##_id, rb_is_##type##_name, message, strlen(message))
 static ID
-check_setter_id(VALUE name, int (*valid_id_p)(ID), int (*valid_name_p)(VALUE),
-		const char *message)
+check_setter_id(VALUE obj, VALUE *pname,
+		int (*valid_id_p)(ID), int (*valid_name_p)(VALUE),
+		const char *message, size_t message_len)
 {
-    ID id;
-    if (SYMBOL_P(name)) {
-	id = SYM2ID(name);
-	if (!valid_id_p(id)) {
-	    rb_name_error(id, message, QUOTE_ID(id));
-	}
-    }
-    else {
-	VALUE str = rb_check_string_type(name);
-	if (NIL_P(str)) {
-	    rb_raise(rb_eTypeError, "%+"PRIsVALUE" is not a symbol or string",
-		     str);
-	}
-	if (!valid_name_p(str)) {
-	    rb_name_error_str(str, message, QUOTE(str));
-	}
-	id = rb_to_id(str);
+    ID id = rb_check_id(pname);
+    VALUE name = *pname;
+
+    if (id ? !valid_id_p(id) : !valid_name_p(name)) {
+	rb_name_err_raise_str(rb_fstring_new(message, message_len),
+			      obj, name);
     }
     return id;
 }
 
-static int
-rb_is_attr_id(ID id)
-{
-    return rb_is_local_id(id) || rb_is_const_id(id);
-}
-
 static int
 rb_is_attr_name(VALUE name)
 {
     return rb_is_local_name(name) || rb_is_const_name(name);
 }
 
-static const char invalid_attribute_name[] = "invalid attribute name `%"PRIsVALUE"'";
-
-static ID
-id_for_attr(VALUE name)
+static int
+rb_is_attr_id(ID id)
 {
-    return id_for_setter(name, attr, invalid_attribute_name);
+    return rb_is_local_id(id) || rb_is_const_id(id);
 }
 
-ID
-rb_check_attr_id(ID id)
+static const char wrong_constant_name[] = "wrong constant name %1$s";
+static const char invalid_attribute_name[] = "invalid attribute name `%1$s'";
+
+static ID
+id_for_attr(VALUE obj, VALUE name)
 {
-    if (!rb_is_attr_id(id)) {
-	rb_name_error_str(id, invalid_attribute_name, QUOTE_ID(id));
-    }
+    ID id = id_for_setter(obj, name, attr, invalid_attribute_name);
+    if (!id) id = rb_intern_str(name);
     return id;
 }
 
@@ -1976,7 +1962,7 @@ rb_mod_attr_reader(int argc, VALUE *argv, VALUE klass)
     int i;
 
     for (i=0; i<argc; i++) {
-	rb_attr(klass, id_for_attr(argv[i]), TRUE, FALSE, TRUE);
+	rb_attr(klass, id_for_attr(klass, argv[i]), TRUE, FALSE, TRUE);
     }
     return Qnil;
 }
@@ -1986,7 +1972,7 @@ rb_mod_attr(int argc, VALUE *argv, VALUE klass)
 {
     if (argc == 2 && (argv[1] == Qtrue || argv[1] == Qfalse)) {
 	rb_warning("optional boolean argument is obsoleted");
-	rb_attr(klass, id_for_attr(argv[0]), 1, RTEST(argv[1]), TRUE);
+	rb_attr(klass, id_for_attr(klass, argv[0]), 1, RTEST(argv[1]), TRUE);
 	return Qnil;
     }
     return rb_mod_attr_reader(argc, argv, klass);
@@ -2008,7 +1994,7 @@ rb_mod_attr_writer(int argc, VALUE *argv, VALUE klass)
     int i;
 
     for (i=0; i<argc; i++) {
-	rb_attr(klass, id_for_attr(argv[i]), FALSE, TRUE, TRUE);
+	rb_attr(klass, id_for_attr(klass, argv[i]), FALSE, TRUE, TRUE);
     }
     return Qnil;
 }
@@ -2036,7 +2022,7 @@ rb_mod_attr_accessor(int argc, VALUE *argv, VALUE klass)
     int i;
 
     for (i=0; i<argc; i++) {
-	rb_attr(klass, id_for_attr(argv[i]), TRUE, TRUE, TRUE);
+	rb_attr(klass, id_for_attr(klass, argv[i]), TRUE, TRUE, TRUE);
     }
     return Qnil;
 }
@@ -2089,17 +2075,14 @@ rb_mod_const_get(int argc, VALUE *argv, VALUE mod)
     const char *pbeg, *p, *path, *pend;
     ID id;
 
-    if (argc == 1) {
-	name = argv[0];
-	recur = Qtrue;
-    }
-    else {
-	rb_scan_args(argc, argv, "11", &name, &recur);
-    }
+    rb_check_arity(argc, 1, 2);
+    name = argv[0];
+    recur = (argc == 1) ? Qtrue : argv[1];
 
     if (SYMBOL_P(name)) {
-	id = SYM2ID(name);
-	if (!rb_is_const_id(id)) goto wrong_id;
+	if (!rb_is_const_sym(name)) goto wrong_name;
+	id = rb_check_id(&name);
+	if (!id) return rb_const_missing(mod, name);
 	return RTEST(recur) ? rb_const_get(mod, id) : rb_const_get_at(mod, id);
     }
 
@@ -2115,8 +2098,7 @@ rb_mod_const_get(int argc, VALUE *argv, VALUE mod)
 
     if (p >= pend || !*p) {
       wrong_name:
-	rb_raise(rb_eNameError, "wrong constant name %"PRIsVALUE,
-		 QUOTE(name));
+	rb_name_err_raise(wrong_constant_name, mod, name);
     }
 
     if (p + 2 < pend && p[0] == ':' && p[1] == ':') {
@@ -2151,22 +2133,21 @@ rb_mod_const_get(int argc, VALUE *argv, VALUE mod)
 	    part = rb_str_subseq(name, beglen, len);
 	    OBJ_FREEZE(part);
 	    if (!ISUPPER(*pbeg) || !rb_is_const_name(part)) {
-		rb_name_error_str(part, "wrong constant name %"PRIsVALUE,
-				  QUOTE(part));
+		name = part;
+		goto wrong_name;
 	    }
 	    else if (!rb_method_basic_definition_p(CLASS_OF(mod), id_const_missing)) {
-		id = rb_intern_str(part);
+		part = rb_str_intern(part);
+		mod = rb_const_missing(mod, part);
+		continue;
 	    }
 	    else {
-		rb_name_error_str(part, "uninitialized constant %"PRIsVALUE"%"PRIsVALUE,
-				  rb_str_subseq(name, 0, beglen),
-				  QUOTE(part));
+		rb_mod_const_missing(mod, part);
 	    }
 	}
 	if (!rb_is_const_id(id)) {
-	  wrong_id:
-	    rb_name_error(id, "wrong constant name %"PRIsVALUE,
-			  QUOTE_ID(id));
+	    name = ID2SYM(id);
+	    goto wrong_name;
 	}
 	mod = RTEST(recur) ? rb_const_get(mod, id) : rb_const_get_at(mod, id);
     }
@@ -2196,7 +2177,8 @@ rb_mod_const_get(int argc, VALUE *argv, VALUE mod)
 static VALUE
 rb_mod_const_set(VALUE mod, VALUE name, VALUE value)
 {
-    ID id = id_for_setter(name, const, "wrong constant name %"PRIsVALUE);
+    ID id = id_for_setter(mod, name, const, wrong_constant_name);
+    if (!id) id = rb_intern_str(name);
     rb_const_set(mod, id, value);
     return value;
 }
@@ -2250,17 +2232,14 @@ rb_mod_const_defined(int argc, VALUE *argv, VALUE mod)
     const char *pbeg, *p, *path, *pend;
     ID id;
 
-    if (argc == 1) {
-	name = argv[0];
-	recur = Qtrue;
-    }
-    else {
-	rb_scan_args(argc, argv, "11", &name, &recur);
-    }
+    rb_check_arity(argc, 1, 2);
+    name = argv[0];
+    recur = (argc == 1) ? Qtrue : argv[1];
 
     if (SYMBOL_P(name)) {
-	id = SYM2ID(name);
-	if (!rb_is_const_id(id)) goto wrong_id;
+	if (!rb_is_const_sym(name)) goto wrong_name;
+	id = rb_check_id(&name);
+	if (!id) return Qfalse;
 	return RTEST(recur) ? rb_const_defined(mod, id) : rb_const_defined_at(mod, id);
     }
 
@@ -2276,8 +2255,7 @@ rb_mod_const_defined(int argc, VALUE *argv, VALUE mod)
 
     if (p >= pend || !*p) {
       wrong_name:
-	rb_raise(rb_eNameError, "wrong constant name %"PRIsVALUE,
-		 QUOTE(name));
+	rb_name_err_raise(wrong_constant_name, mod, name);
     }
 
     if (p + 2 < pend && p[0] == ':' && p[1] == ':') {
@@ -2307,17 +2285,16 @@ rb_mod_const_defined(int argc, VALUE *argv, VALUE mod)
 	    part = rb_str_subseq(name, beglen, len);
 	    OBJ_FREEZE(part);
 	    if (!ISUPPER(*pbeg) || !rb_is_const_name(part)) {
-		rb_name_error_str(part, "wrong constant name %"PRIsVALUE,
-				  QUOTE(part));
+		name = part;
+		goto wrong_name;
 	    }
 	    else {
 		return Qfalse;
 	    }
 	}
 	if (!rb_is_const_id(id)) {
-	  wrong_id:
-	    rb_name_error(id, "wrong constant name %"PRIsVALUE,
-			  QUOTE_ID(id));
+	    name = ID2SYM(id);
+	    goto wrong_name;
 	}
 	if (RTEST(recur)) {
 	    if (!rb_const_defined(mod, id))
@@ -2365,20 +2342,10 @@ rb_mod_const_defined(int argc, VALUE *argv, VALUE mod)
 static VALUE
 rb_obj_ivar_get(VALUE obj, VALUE iv)
 {
-    ID id = rb_check_id(&iv);
+    ID id = id_for_var(obj, iv, an, instance);
 
     if (!id) {
-	if (rb_is_instance_name(iv)) {
-	    return Qnil;
-	}
-	else {
-	    rb_name_error_str(iv, "`%"PRIsVALUE"' is not allowed as an instance variable name",
-			      QUOTE(iv));
-	}
-    }
-    if (!rb_is_instance_id(id)) {
-	rb_name_error(id, "`%"PRIsVALUE"' is not allowed as an instance variable name",
-		      QUOTE_ID(id));
+	return Qnil;
     }
     return rb_ivar_get(obj, id);
 }
@@ -2409,7 +2376,8 @@ rb_obj_ivar_get(VALUE obj, VALUE iv)
 static VALUE
 rb_obj_ivar_set(VALUE obj, VALUE iv, VALUE val)
 {
-    ID id = id_for_setter(iv, instance, "`%"PRIsVALUE"' is not allowed as an instance variable name");
+    ID id = id_for_var(obj, iv, an, instance);
+    if (!id) id = rb_intern_str(iv);
     return rb_ivar_set(obj, id, val);
 }
 
@@ -2436,20 +2404,10 @@ rb_obj_ivar_set(VALUE obj, VALUE iv, VALUE val)
 static VALUE
 rb_obj_ivar_defined(VALUE obj, VALUE iv)
 {
-    ID id = rb_check_id(&iv);
+    ID id = id_for_var(obj, iv, an, instance);
 
     if (!id) {
-	if (rb_is_instance_name(iv)) {
-	    return Qfalse;
-	}
-	else {
-	    rb_name_error_str(iv, "`%"PRIsVALUE"' is not allowed as an instance variable name",
-			      QUOTE(iv));
-	}
-    }
-    if (!rb_is_instance_id(id)) {
-	rb_name_error(id, "`%"PRIsVALUE"' is not allowed as an instance variable name",
-		      QUOTE_ID(id));
+	return Qfalse;
     }
     return rb_ivar_defined(obj, id);
 }
@@ -2473,21 +2431,11 @@ rb_obj_ivar_defined(VALUE obj, VALUE iv)
 static VALUE
 rb_mod_cvar_get(VALUE obj, VALUE iv)
 {
-    ID id = rb_check_id(&iv);
+    ID id = id_for_var(obj, iv, a, class);
 
     if (!id) {
-	if (rb_is_class_name(iv)) {
-	    rb_name_error_str(iv, "uninitialized class variable %"PRIsVALUE" in %"PRIsVALUE"",
-			      iv, rb_class_name(obj));
-	}
-	else {
-	    rb_name_error_str(iv, "`%"PRIsVALUE"' is not allowed as a class variable name",
-			      QUOTE(iv));
-	}
-    }
-    if (!rb_is_class_id(id)) {
-	rb_name_error(id, "`%"PRIsVALUE"' is not allowed as a class variable name",
-		      QUOTE_ID(id));
+	rb_name_err_raise("uninitialized class variable %1$s in %2$s",
+			  obj, iv);
     }
     return rb_cvar_get(obj, id);
 }
@@ -2515,7 +2463,8 @@ rb_mod_cvar_get(VALUE obj, VALUE iv)
 static VALUE
 rb_mod_cvar_set(VALUE obj, VALUE iv, VALUE val)
 {
-    ID id = id_for_setter(iv, class, "`%"PRIsVALUE"' is not allowed as a class variable name");
+    ID id = id_for_var(obj, iv, a, class);
+    if (!id) id = rb_intern_str(iv);
     rb_cvar_set(obj, id, val);
     return val;
 }
@@ -2539,20 +2488,10 @@ rb_mod_cvar_set(VALUE obj, VALUE iv, VALUE val)
 static VALUE
 rb_mod_cvar_defined(VALUE obj, VALUE iv)
 {
-    ID id = rb_check_id(&iv);
+    ID id = id_for_var(obj, iv, a, class);
 
     if (!id) {
-	if (rb_is_class_name(iv)) {
-	    return Qfalse;
-	}
-	else {
-	    rb_name_error_str(iv, "`%"PRIsVALUE"' is not allowed as a class variable name",
-			      QUOTE(iv));
-	}
-    }
-    if (!rb_is_class_id(id)) {
-	rb_name_error(id, "`%"PRIsVALUE"' is not allowed as a class variable name",
-		      QUOTE_ID(id));
+	return Qfalse;
     }
     return rb_cvar_defined(obj, id);
 }
@@ -2578,20 +2517,22 @@ rb_mod_singleton_p(VALUE klass)
     return Qfalse;
 }
 
-static struct conv_method_tbl {
-    const char *method;
-    ID id;
+static const struct conv_method_tbl {
+    const char method[6];
+    unsigned short id;
 } conv_method_names[] = {
-    {"to_int", 0},
-    {"to_ary", 0},
-    {"to_str", 0},
-    {"to_sym", 0},
-    {"to_hash", 0},
-    {"to_proc", 0},
-    {"to_io", 0},
-    {"to_a", 0},
-    {"to_s", 0},
-    {NULL, 0}
+#define M(n) {#n, (unsigned short)idTo_##n}
+    M(int),
+    M(ary),
+    M(str),
+    M(sym),
+    M(hash),
+    M(proc),
+    M(io),
+    M(a),
+    M(s),
+    M(i),
+#undef M
 };
 #define IMPLICIT_CONVERSIONS 7
 
@@ -2599,27 +2540,34 @@ static VALUE
 convert_type(VALUE val, const char *tname, const char *method, int raise)
 {
     ID m = 0;
-    int i;
+    int i = numberof(conv_method_names);
     VALUE r;
+    static const char prefix[] = "to_";
 
-    for (i=0; conv_method_names[i].method; i++) {
-	if (conv_method_names[i].method[0] == method[0] &&
-	    strcmp(conv_method_names[i].method, method) == 0) {
-	    m = conv_method_names[i].id;
-	    break;
+    if (strncmp(prefix, method, sizeof(prefix)-1) == 0) {
+	const char *const meth = &method[sizeof(prefix)-1];
+	for (i=0; i < numberof(conv_method_names); i++) {
+	    if (conv_method_names[i].method[0] == meth[0] &&
+		strcmp(conv_method_names[i].method, meth) == 0) {
+		m = conv_method_names[i].id;
+		break;
+	    }
 	}
     }
     if (!m) m = rb_intern(method);
     r = rb_check_funcall(val, m, 0, 0);
     if (r == Qundef) {
 	if (raise) {
-	    rb_raise(rb_eTypeError, i < IMPLICIT_CONVERSIONS
-                ? "no implicit conversion of %s into %s"
-                : "can't convert %s into %s",
-		     NIL_P(val) ? "nil" :
-		     val == Qtrue ? "true" :
-		     val == Qfalse ? "false" :
-		     rb_obj_classname(val),
+	    const char *msg = i < IMPLICIT_CONVERSIONS ?
+		"no implicit conversion of" : "can't convert";
+	    const char *cname = NIL_P(val) ? "nil" :
+		val == Qtrue ? "true" :
+		val == Qfalse ? "false" :
+		NULL;
+	    if (cname)
+		rb_raise(rb_eTypeError, "%s %s into %s", msg, cname, tname);
+	    rb_raise(rb_eTypeError, "%s %"PRIsVALUE" into %s", msg,
+		     rb_obj_class(val),
 		     tname);
 	}
 	return Qnil;
@@ -2627,6 +2575,16 @@ convert_type(VALUE val, const char *tname, const char *method, int raise)
     return r;
 }
 
+NORETURN(static void conversion_mismatch(VALUE, const char *, const char *, VALUE));
+static void
+conversion_mismatch(VALUE val, const char *tname, const char *method, VALUE result)
+{
+    VALUE cname = rb_obj_class(val);
+    rb_raise(rb_eTypeError,
+	     "can't convert %"PRIsVALUE" to %s (%"PRIsVALUE"#%s gives %"PRIsVALUE")",
+	     cname, tname, cname, method, rb_obj_class(result));
+}
+
 VALUE
 rb_convert_type(VALUE val, int type, const char *tname, const char *method)
 {
@@ -2635,9 +2593,7 @@ rb_convert_type(VALUE val, int type, const char *tname, const char *method)
     if (TYPE(val) == type) return val;
     v = convert_type(val, tname, method, TRUE);
     if (TYPE(v) != type) {
-	const char *cname = rb_obj_classname(val);
-	rb_raise(rb_eTypeError, "can't convert %s to %s (%s#%s gives %s)",
-		 cname, tname, cname, method, rb_obj_classname(v));
+	conversion_mismatch(val, tname, method, v);
     }
     return v;
 }
@@ -2652,9 +2608,7 @@ rb_check_convert_type(VALUE val, int type, const char *tname, const char *method
     v = convert_type(val, tname, method, FALSE);
     if (NIL_P(v)) return Qnil;
     if (TYPE(v) != type) {
-	const char *cname = rb_obj_classname(val);
-	rb_raise(rb_eTypeError, "can't convert %s to %s (%s#%s gives %s)",
-		 cname, tname, cname, method, rb_obj_classname(v));
+	conversion_mismatch(val, tname, method, v);
     }
     return v;
 }
@@ -2669,9 +2623,7 @@ rb_to_integer(VALUE val, const char *method)
     if (RB_TYPE_P(val, T_BIGNUM)) return val;
     v = convert_type(val, "Integer", method, TRUE);
     if (!rb_obj_is_kind_of(v, rb_cInteger)) {
-	const char *cname = rb_obj_classname(val);
-	rb_raise(rb_eTypeError, "can't convert %s to Integer (%s#%s gives %s)",
-		 cname, cname, method, rb_obj_classname(v));
+	conversion_mismatch(val, "Integer", method, v);
     }
     return v;
 }
@@ -2908,33 +2860,84 @@ rb_str_to_dbl(VALUE str, int badcheck)
     return ret;
 }
 
+#define fix2dbl_without_to_f(x) (double)FIX2LONG(x)
+#define big2dbl_without_to_f(x) rb_big2dbl(x)
+#define int2dbl_without_to_f(x) \
+    (FIXNUM_P(x) ? fix2dbl_without_to_f(x) : big2dbl_without_to_f(x))
+#define rat2dbl_without_to_f(x) \
+    (int2dbl_without_to_f(rb_rational_num(x)) / \
+     int2dbl_without_to_f(rb_rational_den(x)))
+
+#define special_const_to_float(val, pre, post) \
+    switch (val) { \
+      case Qnil: \
+	rb_raise(rb_eTypeError, pre "nil" post); \
+      case Qtrue: \
+	rb_raise(rb_eTypeError, pre "true" post); \
+      case Qfalse: \
+	rb_raise(rb_eTypeError, pre "false" post); \
+    }
+
+static inline void
+conversion_to_float(VALUE val)
+{
+    special_const_to_float(val, "can't convert ", " into Float");
+}
+
+static inline void
+implicit_conversion_to_float(VALUE val)
+{
+    special_const_to_float(val, "no implicit conversion to float from ", "");
+}
+
+static int
+to_float(VALUE *valp)
+{
+    VALUE val = *valp;
+    if (SPECIAL_CONST_P(val)) {
+	if (FIXNUM_P(val)) {
+	    *valp = DBL2NUM(fix2dbl_without_to_f(val));
+	    return T_FLOAT;
+	}
+	else if (FLONUM_P(val)) {
+	    return T_FLOAT;
+	}
+	else {
+	    conversion_to_float(val);
+	}
+    }
+    else {
+	int type = BUILTIN_TYPE(val);
+	switch (type) {
+	  case T_FLOAT:
+	    return T_FLOAT;
+	  case T_BIGNUM:
+	    *valp = DBL2NUM(big2dbl_without_to_f(val));
+	    return T_FLOAT;
+	  case T_RATIONAL:
+	    *valp = DBL2NUM(rat2dbl_without_to_f(val));
+	    return T_FLOAT;
+	  case T_STRING:
+	    return T_STRING;
+	}
+    }
+    return T_NONE;
+}
+
 VALUE
 rb_Float(VALUE val)
 {
-    switch (TYPE(val)) {
-      case T_FIXNUM:
-	return DBL2NUM((double)FIX2LONG(val));
-
+    switch (to_float(&val)) {
       case T_FLOAT:
 	return val;
-
-      case T_BIGNUM:
-	return DBL2NUM(rb_big2dbl(val));
-
       case T_STRING:
 	return DBL2NUM(rb_str_to_dbl(val, TRUE));
-
-      case T_NIL:
-	rb_raise(rb_eTypeError, "can't convert nil into Float");
-	break;
-
-      default:
-	return rb_convert_type(val, T_FLOAT, "Float", "to_f");
     }
-
-    UNREACHABLE;
+    return rb_convert_type(val, T_FLOAT, "Float", "to_f");
 }
 
+FUNC_MINIMIZED(static VALUE rb_f_float(VALUE obj, VALUE arg));
+
 /*
  *  call-seq:
  *     Float(arg)    -> float
@@ -2953,18 +2956,24 @@ rb_f_float(VALUE obj, VALUE arg)
     return rb_Float(arg);
 }
 
+static VALUE
+numeric_to_float(VALUE val)
+{
+    if (!rb_obj_is_kind_of(val, rb_cNumeric)) {
+	rb_raise(rb_eTypeError, "can't convert %"PRIsVALUE" into Float",
+		 rb_obj_class(val));
+    }
+    return rb_convert_type(val, T_FLOAT, "Float", "to_f");
+}
+
 VALUE
 rb_to_float(VALUE val)
 {
-    if (RB_TYPE_P(val, T_FLOAT)) return val;
-    if (!rb_obj_is_kind_of(val, rb_cNumeric)) {
-	rb_raise(rb_eTypeError, "can't convert %s into Float",
-		 NIL_P(val) ? "nil" :
-		 val == Qtrue ? "true" :
-		 val == Qfalse ? "false" :
-		 rb_obj_classname(val));
+    switch (to_float(&val)) {
+      case T_FLOAT:
+	return val;
     }
-    return rb_convert_type(val, T_FLOAT, "Float", "to_f");
+    return numeric_to_float(val);
 }
 
 VALUE
@@ -2977,26 +2986,75 @@ rb_check_to_float(VALUE val)
     return rb_check_convert_type(val, T_FLOAT, "Float", "to_f");
 }
 
+static ID id_to_f;
+
+static inline int
+basic_to_f_p(VALUE klass)
+{
+    return rb_method_basic_definition_p(klass, id_to_f);
+}
+
+double
+rb_num_to_dbl(VALUE val)
+{
+    if (SPECIAL_CONST_P(val)) {
+	if (FIXNUM_P(val)) {
+	    if (basic_to_f_p(rb_cFixnum))
+		return fix2dbl_without_to_f(val);
+	}
+	else if (FLONUM_P(val)) {
+	    return rb_float_flonum_value(val);
+	}
+	else {
+	    conversion_to_float(val);
+	}
+    }
+    else {
+	switch (BUILTIN_TYPE(val)) {
+	  case T_FLOAT:
+	    return rb_float_noflonum_value(val);
+	  case T_BIGNUM:
+	    if (basic_to_f_p(rb_cBignum))
+		return big2dbl_without_to_f(val);
+	    break;
+	  case T_RATIONAL:
+	    if (basic_to_f_p(rb_cRational))
+		return rat2dbl_without_to_f(val);
+	    break;
+	}
+    }
+    val = numeric_to_float(val);
+    return RFLOAT_VALUE(val);
+}
+
 double
 rb_num2dbl(VALUE val)
 {
-    switch (TYPE(val)) {
-      case T_FLOAT:
-	return RFLOAT_VALUE(val);
-
-      case T_STRING:
-	rb_raise(rb_eTypeError, "no implicit conversion to float from string");
-	break;
-
-      case T_NIL:
-	rb_raise(rb_eTypeError, "no implicit conversion to float from nil");
-	break;
-
-      default:
-	break;
+    if (SPECIAL_CONST_P(val)) {
+	if (FIXNUM_P(val)) {
+	    return fix2dbl_without_to_f(val);
+	}
+	else if (FLONUM_P(val)) {
+	    return rb_float_flonum_value(val);
+	}
+	else {
+	    implicit_conversion_to_float(val);
+	}
     }
-
-    return RFLOAT_VALUE(rb_Float(val));
+    else {
+	switch (BUILTIN_TYPE(val)) {
+	  case T_FLOAT:
+	    return rb_float_noflonum_value(val);
+	  case T_BIGNUM:
+	    return big2dbl_without_to_f(val);
+	  case T_RATIONAL:
+	    return rat2dbl_without_to_f(val);
+	  case T_STRING:
+	    rb_raise(rb_eTypeError, "no implicit conversion to float from string");
+	}
+    }
+    val = rb_convert_type(val, T_FLOAT, "Float", "to_f");
+    return RFLOAT_VALUE(val);
 }
 
 VALUE
@@ -3094,6 +3152,68 @@ rb_f_hash(VALUE obj, VALUE arg)
     return rb_Hash(arg);
 }
 
+struct dig_method {
+    VALUE klass;
+    int basic;
+};
+
+static ID id_dig;
+
+static int
+dig_basic_p(VALUE obj, struct dig_method *cache)
+{
+    VALUE klass = RBASIC_CLASS(obj);
+    if (klass != cache->klass) {
+	cache->klass = klass;
+	cache->basic = rb_method_basic_definition_p(klass, id_dig);
+    }
+    return cache->basic;
+}
+
+static void
+no_dig_method(int found, VALUE recv, ID mid, int argc, const VALUE *argv, VALUE data)
+{
+    if (!found) {
+	rb_raise(rb_eTypeError, "%"PRIsVALUE" does not have #dig method",
+		 CLASS_OF(data));
+    }
+}
+
+VALUE
+rb_obj_dig(int argc, VALUE *argv, VALUE obj, VALUE notfound)
+{
+    struct dig_method hash = {Qnil}, ary = {Qnil}, strt = {Qnil};
+
+    for (; argc > 0; ++argv, --argc) {
+	if (NIL_P(obj)) return notfound;
+	if (!SPECIAL_CONST_P(obj)) {
+	    switch (BUILTIN_TYPE(obj)) {
+	      case T_HASH:
+		if (dig_basic_p(obj, &hash)) {
+		    obj = rb_hash_aref(obj, *argv);
+		    continue;
+		}
+		break;
+	      case T_ARRAY:
+		if (dig_basic_p(obj, &ary)) {
+		    obj = rb_ary_at(obj, *argv);
+		    continue;
+		}
+		break;
+	      case T_STRUCT:
+		if (dig_basic_p(obj, &strt)) {
+		    obj = rb_struct_lookup(obj, *argv);
+		    continue;
+		}
+		break;
+	    }
+	}
+	return rb_check_funcall_with_hook(obj, id_dig, argc, argv,
+					  no_dig_method, obj);
+    }
+    return obj;
+}
+
 /*
  *  Document-class: Class
  *
@@ -3248,10 +3368,8 @@ rb_f_hash(VALUE obj, VALUE arg)
  */
 
 void
-Init_Object(void)
+InitVM_Object(void)
 {
-    int i;
-
     Init_class_hierarchy();
 
 #if 0
@@ -3310,6 +3428,7 @@ Init_Object(void)
     rb_define_method(rb_mKernel, "singleton_class", rb_obj_singleton_class, 0);
     rb_define_method(rb_mKernel, "clone", rb_obj_clone, 0);
     rb_define_method(rb_mKernel, "dup", rb_obj_dup, 0);
+    rb_define_method(rb_mKernel, "itself", rb_obj_itself, 0);
     rb_define_method(rb_mKernel, "initialize_copy", rb_obj_init_copy, 1);
     rb_define_method(rb_mKernel, "initialize_dup", rb_obj_init_dup_clone, 1);
     rb_define_method(rb_mKernel, "initialize_clone", rb_obj_init_dup_clone, 1);
@@ -3362,6 +3481,7 @@ Init_Object(void)
     rb_define_method(rb_cNilClass, "&", false_and, 1);
     rb_define_method(rb_cNilClass, "|", false_or, 1);
     rb_define_method(rb_cNilClass, "^", false_xor, 1);
+    rb_define_method(rb_cNilClass, "===", rb_equal, 1);
 
     rb_define_method(rb_cNilClass, "nil?", rb_true, 0);
     rb_undef_alloc_func(rb_cNilClass);
@@ -3394,6 +3514,7 @@ Init_Object(void)
 
     rb_define_alloc_func(rb_cModule, rb_module_s_alloc);
     rb_define_method(rb_cModule, "initialize", rb_mod_initialize, 0);
+    rb_define_method(rb_cModule, "initialize_clone", rb_mod_initialize_clone, 1);
     rb_define_method(rb_cModule, "instance_methods", rb_class_instance_methods, -1); /* in class.c */
     rb_define_method(rb_cModule, "public_instance_methods",
 		     rb_class_public_instance_methods, -1);    /* in class.c */
@@ -3419,6 +3540,7 @@ Init_Object(void)
     rb_define_method(rb_cModule, "class_variable_defined?", rb_mod_cvar_defined, 1);
     rb_define_method(rb_cModule, "public_constant", rb_mod_public_constant, -1); /* in variable.c */
     rb_define_method(rb_cModule, "private_constant", rb_mod_private_constant, -1); /* in variable.c */
+    rb_define_method(rb_cModule, "deprecate_constant", rb_mod_deprecate_constant, -1); /* in variable.c */
     rb_define_method(rb_cModule, "singleton_class?", rb_mod_singleton_p, 0);
 
     rb_define_method(rb_cClass, "allocate", rb_obj_alloc, 0);
@@ -3445,6 +3567,7 @@ Init_Object(void)
     rb_define_method(rb_cTrueClass, "&", true_and, 1);
     rb_define_method(rb_cTrueClass, "|", true_or, 1);
     rb_define_method(rb_cTrueClass, "^", true_xor, 1);
+    rb_define_method(rb_cTrueClass, "===", rb_equal, 1);
     rb_undef_alloc_func(rb_cTrueClass);
     rb_undef_method(CLASS_OF(rb_cTrueClass), "new");
     /*
@@ -3458,14 +3581,19 @@ Init_Object(void)
     rb_define_method(rb_cFalseClass, "&", false_and, 1);
     rb_define_method(rb_cFalseClass, "|", false_or, 1);
     rb_define_method(rb_cFalseClass, "^", false_xor, 1);
+    rb_define_method(rb_cFalseClass, "===", rb_equal, 1);
     rb_undef_alloc_func(rb_cFalseClass);
     rb_undef_method(CLASS_OF(rb_cFalseClass), "new");
     /*
      * An alias of +false+
      */
     rb_define_global_const("FALSE", Qfalse);
+}
 
-    for (i=0; conv_method_names[i].method; i++) {
-	conv_method_names[i].id = rb_intern(conv_method_names[i].method);
-    }
+void
+Init_Object(void)
+{
+    id_to_f = rb_intern_const("to_f");
+    id_dig = rb_intern_const("dig");
+    InitVM(Object);
 }
